/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available at through the world-wide-web at                           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Jon Parise <jon@php.net>                                     |
  +----------------------------------------------------------------------+
*/

/* $Id: sasl.c,v 1.3 2004/01/23 08:34:28 jon Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "php_sasl.h"
#include "ext/standard/info.h"

#define SASL_CONSTANT(c) REGISTER_LONG_CONSTANT(#c, c, \
												CONST_CS | CONST_PERSISTENT)

#if SASL_VERSION_MAJOR < 2
typedef char php_sasl_char;
#define PHP_SASL_FREE(p) if (p) free(p);
#else
typedef const char php_sasl_char;
#define PHP_SASL_FREE(p)
#endif

#define le_conn_name "SASL Connection Context"
static int le_conn;

PHPAPI ZEND_DECLARE_MODULE_GLOBALS(sasl);

/* SASL Callback Functions */
/* {{{ php_sasl_cb_getopt
 */
static int php_sasl_cb_getopt(void *context, const char *plugin_name,
							  const char *option, const char **result,
							  unsigned *len)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_STRING(args[0], (plugin_name) ? (char *)plugin_name : "", 1);
	ZVAL_STRING(args[1], (char *)option, 1);

	if (call_user_function(CG(function_table), NULL, function, retval,
					   	   2, args TSRMLS_CC) == SUCCESS) {
		ret = Z_LVAL_P(retval);
		*result = estrdup(Z_STRVAL_P(retval));
		*len = Z_STRLEN_P(retval);
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_log
 */
static int php_sasl_cb_log(void *context, int level, const char *message)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], level);
	ZVAL_STRING(args[1], (char *)message, 1);

	if (call_user_function(CG(function_table), NULL, function, retval,
					   	   2, args TSRMLS_CC) == SUCCESS) {
		ret = Z_LVAL_P(retval);
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_getpath
 */
static int php_sasl_cb_getpath(void *context, const char **path)
{
	zval *function = context;
	zval *retval;
	int ret = SASL_FAIL;

	MAKE_STD_ZVAL(retval);

	if (call_user_function(CG(function_table), NULL, function, retval,
						   0, NULL TSRMLS_CC) == SUCCESS) {
		ret = Z_LVAL_P(retval);
		*path = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&retval);

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_verifyfile
 */
static int php_sasl_cb_verifyfile(void *context, const char *file,
								  sasl_verify_type_t type)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_STRING(args[0], (char *)file, 1);
	ZVAL_LONG(args[1], type);

	if (call_user_function(CG(function_table), NULL, function, retval,
					   	   2, args TSRMLS_CC) == SUCCESS) {
		ret = Z_LVAL_P(retval);
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_getconfpath
 */
static int php_sasl_cb_getconfpath(void *context, char **path)
{
	zval *function = context;
	zval *retval;
	int ret = SASL_FAIL;

	MAKE_STD_ZVAL(retval);

	if (call_user_function(CG(function_table), NULL, function, retval,
						   0, NULL TSRMLS_CC) == SUCCESS) {
		ret = Z_LVAL_P(retval);
		*path = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&retval);

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_getsimple
 */
static int php_sasl_cb_getsimple(void *context, int id, const char **result,
								 unsigned *len)
{
	zval *function = context;
	zval *args[1];
	zval *retval;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);

	if (call_user_function(CG(function_table), NULL, function, retval,
						   1, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
		*len = Z_STRLEN_P(retval);
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&retval);

	return SASL_OK;
}
/* }}} */
/* {{{ php_sasl_cb_getsecret
 */
static int php_sasl_cb_getsecret(sasl_conn_t *conn, void *context, int id,
								 sasl_secret_t **psecret)
{
	zval *function = context;
	zval *args[2];
	zval *retval;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_NULL(args[0]);
	//ZVAL_RESOURCE(args[0], r);
	ZVAL_LONG(args[1], id);

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		int len = Z_STRLEN_P(retval);
		*psecret = (sasl_secret_t *)emalloc(sizeof(sasl_secret_t) +
											(len * sizeof(char)));
		if (*psecret) {
			(*psecret)->len = len;
			memcpy((*psecret)->data, Z_STRVAL_P(retval), len);
		}
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);

	return SASL_OK;
}
/* }}} */
/* {{{ php_sasl_cb_chalprompt
 */
static int php_sasl_cb_chalprompt(void *context, int id, const char *challenge,
								  const char *prompt, const char *defresult,
								  const char **result, unsigned *len)
{
	zval *function = context;
	zval *args[4];
	zval *retval;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(args[2]);
	MAKE_STD_ZVAL(args[3]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_STRING(args[1], (char *)challenge, 1);
	ZVAL_STRING(args[2], (char *)prompt, 1);
	ZVAL_STRING(args[3], (char *)defresult, 1);

	if (call_user_function(CG(function_table), NULL, function, retval,
						   4, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
		*len = Z_STRLEN_P(retval);
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&args[2]);
	zval_ptr_dtor(&args[3]);
	zval_ptr_dtor(&retval);

	return SASL_OK;
}
/* }}} */
/* {{{ php_sasl_cb_getrealm
 */
static int php_sasl_cb_getrealm(void *context, int id,
								const char **availrealms, const char **result)
{
	zval *function = context;
	zval *args[2];
	zval *retval;

	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_NULL(args[1]); // TODO: Make an array of available realms

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);

	return SASL_OK;
}
/* }}} */
/* {{{ php_sasl_cb_authorize
 */
static int php_sasl_cb_authorize(sasl_conn_t *conn, void *context,
								 const char *requested_user, unsigned rlen,
								 const char *auth_identity, unsigned alen,
								 const char *def_realm, unsigned urlen,
								 struct propctx *propctx)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_NOAUTHZ;

#if 0
	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_NULL(args[1]); // TODO: Make an array of available realms

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);
#endif

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_userdb_checkpass
 */
static int php_sasl_cb_userdb_checkpass(sasl_conn_t *conn, void *context,
										const char *user, const char *pass,
										unsigned passlen,
										struct propctx *propctx)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

#if 0
	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_NULL(args[1]); // TODO: Make an array of available realms

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);
#endif

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_userdb_setpass
 */
static int php_sasl_cb_userdb_setpass(sasl_conn_t *conn, void *context,
									  const char *user, const char *pass,
									  unsigned passlen,
									  struct propctx *propctx, unsigned flags)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

#if 0
	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_NULL(args[1]); // TODO: Make an array of available realms

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);
#endif

	return ret;
}
/* }}} */
/* {{{ php_sasl_cb_canon_user
 */
static int php_sasl_cb_canon_user(sasl_conn_t *conn, void *context,
								  const char *in, unsigned inlen,
								  unsigned flags, const char *user_realm,
								  char *out, unsigned out_max,
								  unsigned *out_len)
{
	zval *function = context;
	zval *args[2];
	zval *retval;
	int ret = SASL_FAIL;

#if 0
	MAKE_STD_ZVAL(args[0]);
	MAKE_STD_ZVAL(args[1]);
	MAKE_STD_ZVAL(retval);

	ZVAL_LONG(args[0], id);
	ZVAL_NULL(args[1]); // TODO: Make an array of available realms

	if (call_user_function(CG(function_table), NULL, function, retval,
						   2, args TSRMLS_CC) == SUCCESS) {
		*result = estrdup(Z_STRVAL_P(retval));
	}

	zval_ptr_dtor(&args[0]);
	zval_ptr_dtor(&args[1]);
	zval_ptr_dtor(&retval);
#endif

	return ret;
}
/* }}} */

/* {{{ sasl_callback_function_t php_sasl_get_callback_function(unsigned long id)
 */
typedef int (*sasl_callback_function_t)();
static sasl_callback_function_t php_sasl_get_callback_function(unsigned long id)
{
	switch (id) {
	case SASL_CB_GETOPT:					return php_sasl_cb_getopt;
	case SASL_CB_LOG:						return php_sasl_cb_log;
	case SASL_CB_GETPATH:					return php_sasl_cb_getpath;
	case SASL_CB_VERIFYFILE:				return php_sasl_cb_verifyfile;
	case SASL_CB_GETCONFPATH:				return php_sasl_cb_getconfpath;
	case SASL_CB_USER:						return php_sasl_cb_getsimple;
	case SASL_CB_AUTHNAME:					return php_sasl_cb_getsimple;
	case SASL_CB_LANGUAGE:					return php_sasl_cb_getsimple;
	case SASL_CB_CNONCE:					return php_sasl_cb_getsimple;
	case SASL_CB_PASS:						return php_sasl_cb_getsecret;
	case SASL_CB_ECHOPROMPT:				return php_sasl_cb_chalprompt;
	case SASL_CB_NOECHOPROMPT:				return php_sasl_cb_chalprompt;
	case SASL_CB_GETREALM:					return php_sasl_cb_getrealm;
	case SASL_CB_PROXY_POLICY:				return php_sasl_cb_authorize;
	case SASL_CB_SERVER_USERDB_CHECKPASS:	return php_sasl_cb_userdb_checkpass;
	case SASL_CB_SERVER_USERDB_SETPASS:		return php_sasl_cb_userdb_setpass;
	case SASL_CB_CANON_USER:				return php_sasl_cb_canon_user;
	}

	return NULL;
}
/* }}} */
/* {{{ php_sasl_create_callbacks(HashTable *ht)
 */
static int php_sasl_create_callbacks(HashTable *ht, sasl_callback_t *callbacks)
{
	int n = 0;
	zval **entry;
	char *string_key;
	long num_key;

	/*
	 * Start by counting the number of legitimate callbacks that exist in
	 * the array.  This allows us to appropriately size our callback array.
	 */
	zend_hash_internal_pointer_reset(ht);
	while (zend_hash_get_current_data(ht, (void **)&entry) == SUCCESS) {
		switch (zend_hash_get_current_key(ht, &string_key, &num_key, 0)) {
			case HASH_KEY_IS_LONG:
				if (php_sasl_get_callback_function(num_key)) {
					n++;
				}
				break;
		}
		zend_hash_move_forward(ht);
	}

	/*
	 * Now that we know how many callbacks we have, allocate the array.  We
	 * include an additional entry for the final SASL_CB_LIST_END sentinel.
	 */
	callbacks = (sasl_callback_t *)emalloc(sizeof(sasl_callback_t) * (n + 1));

	if (callbacks) {
		/* Assign the sentinel to the last entry in the callback array. */
		callbacks[n].id = SASL_CB_LIST_END;

		/*
		 * Now we iterate through the array again, except this time we will
		 * actually create the callback entries.
		 */
		n = 0;

		zend_hash_internal_pointer_reset(ht);
		while (zend_hash_get_current_data(ht, (void **)&entry) == SUCCESS) {
			sasl_callback_t *cb = &callbacks[n];

			switch (zend_hash_get_current_key(ht, &string_key, &num_key, 0)) {
				case HASH_KEY_IS_LONG:
					{
						cb->proc = php_sasl_get_callback_function(num_key);
						if (cb->proc) {
							cb->id = num_key;
							cb->context = *entry;
							n++;
						}
					}
					break;
			}
			zend_hash_move_forward(ht);
		}
	}

	return n;
}
/* }}} */

/* {{{ php_sasl_error(int level, int reason)
 */
static void php_sasl_error(int level, int code TSRMLS_DC)
{
	php_error_docref(NULL TSRMLS_CC, level,
					 "%s", sasl_errstring(code, NULL, NULL));
}
/* }}} */
/* {{{ php_sasl_rinit_globals(TSRMLS_D)
 */
static void php_sasl_rinit_globals(TSRMLS_D)
{
	SASLG(client_callbacks) = NULL;
	SASLG(server_callbacks) = NULL;
}
/* }}} */
/* {{{ php_sasl_rshutdown_globals(TSRMLS_D)
 */
static void php_sasl_rshutdown_globals(TSRMLS_D)
{
	if (SASLG(client_callbacks)) {
		efree(SASLG(client_callbacks));
		SASLG(client_callbacks) = NULL;
	}

	if (SASLG(server_callbacks)) {
		efree(SASLG(server_callbacks));
		SASLG(server_callbacks) = NULL;
	}
}
/* }}} */
/* {{{ php_sasl_destroy_conn(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 */
static void php_sasl_destroy_conn(zend_rsrc_list_entry *rsrc TSRMLS_DC)
{
	sasl_conn_t *conn = (sasl_conn_t *)rsrc->ptr;
	sasl_dispose(&conn);
}
/* }}} */

/* {{{ sasl_functions[]
 */
zend_function_entry sasl_functions[] = {
	/* Common Functions */
	PHP_FE(sasl_version,			NULL)
	PHP_FE(sasl_errstring,			NULL)
#if SASL_VERSION_MAJOR >= 2
	PHP_FE(sasl_errdetail,			NULL)
	PHP_FE(sasl_seterror,			NULL)
#endif
	PHP_FE(sasl_encode,				NULL)
	PHP_FE(sasl_decode,				NULL)

	/* Client Functions */
	PHP_FE(sasl_client_init,		NULL)
	PHP_FE(sasl_client_new,			NULL)
	PHP_FE(sasl_client_start,		NULL)
	PHP_FE(sasl_client_step,		NULL)

	/* Server Functions */
	PHP_FE(sasl_server_init,		NULL)
	PHP_FE(sasl_server_new,			NULL)
	PHP_FE(sasl_server_start,		NULL)
	PHP_FE(sasl_server_step,		NULL)
	PHP_FE(sasl_listmech,			NULL)
	PHP_FE(sasl_checkpass,			NULL)

	{NULL, NULL, NULL}
};
/* }}} */
/* {{{ sasl_module_entry
 */
zend_module_entry sasl_module_entry = {
	STANDARD_MODULE_HEADER,
	"sasl",
	sasl_functions,
	PHP_MINIT(sasl),
	PHP_MSHUTDOWN(sasl),
	PHP_RINIT(sasl),
	PHP_RSHUTDOWN(sasl),
	PHP_MINFO(sasl),
	PHP_SASL_VERSION,
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_SASL
ZEND_GET_MODULE(sasl)
#endif
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(sasl)
{
	le_conn = zend_register_list_destructors_ex(php_sasl_destroy_conn, NULL,
												le_conn_name, module_number);

	/* SASL result codes */
	SASL_CONSTANT(SASL_CONTINUE);
	SASL_CONSTANT(SASL_OK);
	SASL_CONSTANT(SASL_FAIL);
	SASL_CONSTANT(SASL_NOMEM);
	SASL_CONSTANT(SASL_BUFOVER);
	SASL_CONSTANT(SASL_NOMECH);
	SASL_CONSTANT(SASL_BADPROT);
	SASL_CONSTANT(SASL_NOTDONE);
	SASL_CONSTANT(SASL_BADPARAM);
	SASL_CONSTANT(SASL_TRYAGAIN);
	SASL_CONSTANT(SASL_BADMAC);
#if SASL_VERSION_MAJOR >= 2
	SASL_CONSTANT(SASL_NOTINIT);
#endif
	/* Client only: */
	SASL_CONSTANT(SASL_INTERACT);
	SASL_CONSTANT(SASL_BADSERV);
	SASL_CONSTANT(SASL_WRONGMECH);
	/* Server only: */
	SASL_CONSTANT(SASL_BADAUTH);
	SASL_CONSTANT(SASL_NOAUTHZ);
	SASL_CONSTANT(SASL_TOOWEAK);
	SASL_CONSTANT(SASL_ENCRYPT);
	SASL_CONSTANT(SASL_TRANS);
	SASL_CONSTANT(SASL_EXPIRED);
	SASL_CONSTANT(SASL_DISABLED);
	SASL_CONSTANT(SASL_NOUSER);
	SASL_CONSTANT(SASL_BADVERS);
#if SASL_VERSION_MAJOR >= 2
	SASL_CONSTANT(SASL_UNAVAIL);
	SASL_CONSTANT(SASL_NOVERIFY);
#endif

	/* SASL usage flags (for sasl_server_new and sasl_client_new) */
#if SASL_VERSION_MAJOR >= 2
	SASL_CONSTANT(SASL_SUCCESS_DATA);
	SASL_CONSTANT(SASL_NEED_PROXY);
#endif

	/* SASL security flags */
	SASL_CONSTANT(SASL_SEC_NOPLAINTEXT);
	SASL_CONSTANT(SASL_SEC_NOACTIVE);
	SASL_CONSTANT(SASL_SEC_NODICTIONARY);
	SASL_CONSTANT(SASL_SEC_FORWARD_SECRECY);
	SASL_CONSTANT(SASL_SEC_NOANONYMOUS);
	SASL_CONSTANT(SASL_SEC_PASS_CREDENTIALS);
#if SASL_VERSION_MAJOR >= 2
	SASL_CONSTANT(SASL_SEC_MUTUAL_AUTH);
	SASL_CONSTANT(SASL_SEC_MAXIMUM);
#endif

	/* SASL callback identifiers */
	SASL_CONSTANT(SASL_CB_GETOPT);
	SASL_CONSTANT(SASL_CB_LOG);
	SASL_CONSTANT(SASL_CB_GETPATH);
	SASL_CONSTANT(SASL_CB_VERIFYFILE);
	SASL_CONSTANT(SASL_CB_GETCONFPATH);
	SASL_CONSTANT(SASL_CB_USER);
	SASL_CONSTANT(SASL_CB_AUTHNAME);
	SASL_CONSTANT(SASL_CB_LANGUAGE);
	SASL_CONSTANT(SASL_CB_CNONCE);
	SASL_CONSTANT(SASL_CB_PASS);
	SASL_CONSTANT(SASL_CB_ECHOPROMPT);
	SASL_CONSTANT(SASL_CB_NOECHOPROMPT);
	SASL_CONSTANT(SASL_CB_GETREALM);
	SASL_CONSTANT(SASL_CB_PROXY_POLICY);
	SASL_CONSTANT(SASL_CB_SERVER_USERDB_CHECKPASS);
	SASL_CONSTANT(SASL_CB_SERVER_USERDB_SETPASS);
	SASL_CONSTANT(SASL_CB_CANON_USER);

	/* SASL_CB_VERIFYFILE flags */
	SASL_CONSTANT(SASL_VRFY_PLUGIN);
	SASL_CONSTANT(SASL_VRFY_CONF);
	SASL_CONSTANT(SASL_VRFY_PASSWD);
	SASL_CONSTANT(SASL_VRFY_OTHER);

	return SUCCESS;
}
/* }}} */
/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(sasl)
{
	sasl_done();

	return SUCCESS;
}
/* }}} */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(sasl)
{
	php_sasl_rinit_globals(TSRMLS_C);
	return SUCCESS;
}
/* }}} */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(sasl)
{
	php_sasl_rshutdown_globals(TSRMLS_C);
	return SUCCESS;
}
/* }}} */
/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(sasl)
{
	char api_version[16], lib_version[64];
	const char *sasl_implementation = "Unknown";
	int libsasl_version;
	int libsasl_major = SASL_VERSION_MAJOR;
	int libsasl_minor = SASL_VERSION_MINOR;
	int libsasl_step = SASL_VERSION_STEP;

#if SASL_VERSION_MAJOR >= 2
	/* Grab the version information from the SASL library. */
	sasl_version(&sasl_implementation, &libsasl_version);
	libsasl_major = libsasl_version >> 24;
	libsasl_minor = (libsasl_version >> 16) & 0xFF;
	libsasl_step = libsasl_version & 0xFFFF;
#endif

	/* Build the version strings. */
	snprintf(api_version, 16, "%u.%u.%u",
			 SASL_VERSION_MAJOR, SASL_VERSION_MINOR, SASL_VERSION_STEP);
	snprintf(lib_version, 64, "%u.%u.%u (%s)",
			 libsasl_major, libsasl_minor, libsasl_step, sasl_implementation);

	php_info_print_table_start();
	php_info_print_table_row(2, "SASL Support", "enabled");
	php_info_print_table_row(2, "SASL API Version", api_version);
	php_info_print_table_row(2, "SASL Library Version", lib_version);
	php_info_print_table_row(2, "Extension Version", PHP_SASL_VERSION);
	php_info_print_table_row(2, "CVS Revision", "$Revision$");
	php_info_print_table_end();
}
/* }}} */

/* Common Functions */
/* {{{ proto string sasl_version()
   Return the SASL library version information. */
PHP_FUNCTION(sasl_version)
{
	char version[64];
	const char *sasl_implementation = "Unknown";
	int libsasl_version;
	int libsasl_major = SASL_VERSION_MAJOR;
	int libsasl_minor = SASL_VERSION_MINOR;
	int libsasl_step = SASL_VERSION_STEP;

#if SASL_VERSION_MAJOR >= 2
	/* Grab the version information from the SASL library. */
	sasl_version(&sasl_implementation, &libsasl_version);
	libsasl_major = libsasl_version >> 24;
	libsasl_minor = (libsasl_version >> 16) & 0xFF;
	libsasl_step = libsasl_version & 0xFFFF;
#endif

	snprintf(version, sizeof(version), "%u.%u.%u (%s)",
			 libsasl_major, libsasl_minor, libsasl_step, sasl_implementation);

	RETURN_STRING(version, 1);
}
/* }}} */
/* {{{ proto string sasl_errdetail(resource conn)
   Returns detail about the last error that occured on the given connection. */
#if SASL_VERSION_MAJOR >= 2
PHP_FUNCTION(sasl_errdetail)
{
	zval *rsrc;
	sasl_conn_t *conn;

	if (zend_parse_parameters(1 TSRMLS_CC, "r", &rsrc) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	RETURN_STRING((char *)sasl_errdetail(conn), 1);
}
#endif
/* }}} */
/* {{{ proto string sasl_errstring(int code [, string languages])
   Returns the string translation of the given error code. */
PHP_FUNCTION(sasl_errstring)
{
	long code;
	char *languages = NULL;
	int languages_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l|s",
							  &code, &languages, &languages_len) == FAILURE) {
		return;
	}

	RETURN_STRING((char *)sasl_errstring(code, languages, NULL), 1);
}
/* }}} */
/* {{{ proto void sasl_seterror(resource conn, string message)
   Set the error string which will be returned by sasl_errdetail. */
#if SASL_VERSION_MAJOR >= 2
PHP_FUNCTION(sasl_seterror)
{
	zval *rsrc;
	sasl_conn_t *conn;
	char *message;
	int message_len;

	if (zend_parse_parameters(2 TSRMLS_CC, "rs",
							  &rsrc, &message, &message_len) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	sasl_seterror(conn, 0, message);
}
#endif
/* }}} */
/* {{{ proto string sasl_encode(resource conn, string input)
   Encodes a block of data for tranmission using the security layer. */
PHP_FUNCTION(sasl_encode)
{
	zval *rsrc;
	sasl_conn_t *conn;
	char *input;
	int input_len;
	php_sasl_char *output = NULL;
	unsigned output_len = 0;
	int r = SASL_FAIL;

	if (zend_parse_parameters(2 TSRMLS_CC, "rs",
							  &rsrc, &input, &input_len) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	r = sasl_encode(conn, input, input_len, &output, &output_len);
	if (r != SASL_OK) {
		php_sasl_error(E_NOTICE, r TSRMLS_CC);
		PHP_SASL_FREE(output);
		RETURN_EMPTY_STRING();
	}

	RETVAL_STRINGL((char *)output, output_len, 1);
	PHP_SASL_FREE(output);
}
/* }}} */
/* {{{ proto string sasl_decode(resource conn, string input)
   Decode a block of data received using the security layer. */
PHP_FUNCTION(sasl_decode)
{
	zval *rsrc;
	sasl_conn_t *conn;
	char *input;
	int input_len;
	php_sasl_char *output = NULL;
	unsigned output_len = 0;
	int r = SASL_FAIL;

	if (zend_parse_parameters(2 TSRMLS_CC, "rs",
							  &rsrc, &input, &input_len) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	r = sasl_decode(conn, input, input_len, &output, &output_len);
	if (r != SASL_OK) {
		php_sasl_error(E_NOTICE, r TSRMLS_CC);
		PHP_SASL_FREE(output);
		RETURN_EMPTY_STRING();
	}

	RETVAL_STRINGL((char *)output, output_len, 1);
	PHP_SASL_FREE(output);
}
/* }}} */

/* Client Functions */
/* {{{ proto bool sasl_client_init()
   Initializes the SASL client drivers. */
PHP_FUNCTION(sasl_client_init)
{
	zval *array;

	if (zend_parse_parameters(1 TSRMLS_CC, "a", &array) == FAILURE) {
		return;
	}

	/*
	 * Convert the supplied array into a set of callback functions.  We
	 * don't bother spending any time validating its contents because the
	 * sasl_client_init() function will do that for us (and return an error
	 * if it encounters something that it doesn't like).
	 */
	php_sasl_create_callbacks(Z_ARRVAL_P(array), SASLG(client_callbacks));

	if (sasl_client_init(SASLG(client_callbacks)) != SASL_OK) {
		/* XXX: Free the callbacks? */
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */
/* {{{ proto resource sasl_client_new(string service [, string hostname [, int flags]])
   Creates a new SASL client connection context. */
PHP_FUNCTION(sasl_client_new)
{
	char *service, *hostname = NULL;
	int server_len, hostname_len;
	long flags = 0;
	sasl_conn_t *conn = NULL;
	int r = SASL_FAIL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|sl",
							  &service, &server_len,
							  &hostname, &hostname_len,
							  &flags) == FAILURE) {
		return;
	}

#if SASL_VERSION_MAJOR < 2
	r = sasl_client_new(service, hostname, NULL, flags, &conn);
#else
	r = sasl_client_new(service, hostname, NULL, NULL, NULL, flags, &conn);
#endif

	/* If sasl_client_new() failed, print an error and return false. */
	if (r != SASL_OK) {
		php_sasl_error(E_ERROR, r TSRMLS_CC);
		RETURN_FALSE;
	}

	ZEND_REGISTER_RESOURCE(return_value, conn, le_conn);
}
/* }}} */
/* {{{ proto int sasl_client_start(resource conn, string mechlist [, string &$output [, string &$mech]])
   Starts an authentication session. */
PHP_FUNCTION(sasl_client_start)
{
	zval *rsrc, *output = NULL, *mech = NULL;
	sasl_conn_t *conn;
	char *mechlist;
	int mechlist_len;
	php_sasl_char *data = NULL;
	unsigned data_len = 0;
	const char *chosenmech = NULL;
	int r = SASL_FAIL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|z/z/",
							  &rsrc,
							  &mechlist, &mechlist_len,
							  &output, &mech) == FAILURE) {
		return;
	}

	if (output) {
		zval_dtor(output);
	}

	if (mech) {
		zval_dtor(mech);
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

#if SASL_VERSION_MAJOR < 2
	r = sasl_client_start(conn, mechlist, NULL, NULL, &data, &data_len,
						  &chosenmech);
#else
	r = sasl_client_start(conn, mechlist, NULL, &data, &data_len, &chosenmech);
#endif

	/*
	 * Issue a warning and return the error code if we receive an unexpected
	 * result code.
	 */
	if ((r != SASL_OK) && (r != SASL_CONTINUE)) {
		php_sasl_error(E_WARNING, r TSRMLS_CC);
		RETURN_LONG(r);
	}

	/* Store the output in the "output" parameter (by reference). */
	ZVAL_STRINGL(output, (char *)data, data_len, 1);
	PHP_SASL_FREE(data);

	/* Store the chosen mechanism in the "mech" parameter (by reference). */
	ZVAL_STRING(mech, (char *)chosenmech, 1);

	RETURN_LONG(r);
}
/* }}} */
/* {{{ proto int sasl_client_step(resource conn, string input, string &$output)
   Performs a step in the authentication process. */
PHP_FUNCTION(sasl_client_step)
{
	zval *rsrc, *output = NULL;
	sasl_conn_t *conn;
	char *input;
	int input_len;
	php_sasl_char *data = NULL;
	unsigned data_len = 0;
	int r = SASL_FAIL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsz/",
							  &rsrc,
							  &input, &input_len,
							  &output) == FAILURE) {
		return;
	}

	if (output) {
		zval_dtor(output);
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	r = sasl_client_step(conn, input, input_len, NULL, &data, &data_len);

	/* Print a warning if we receive an unexpected result. */
	if ((r != SASL_OK) && (r != SASL_CONTINUE)) {
		php_sasl_error(r, E_WARNING TSRMLS_CC);
	}

	/* Store the output in the "output" parameter (passed by reference). */
	ZVAL_STRINGL(output, (char *)data, data_len, 1);
	PHP_SASL_FREE(data);

	/* Simply return the result code from sasl_server_step(). */
	RETURN_LONG(r);
}
/* }}} */

/* Server Functions */
/* {{{ proto bool sasl_server_init(string name)
   Initializes the session and loads the shared authentication mechanisms. */
PHP_FUNCTION(sasl_server_init)
{
	zval *array;
	char *name;
	int name_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "as",
							  &array, &name, &name_len) == FAILURE) {
		return;
	}

	/*
	 * Convert the supplied array into a set of callback functions.  We
	 * don't bother spending any time validating its contents because the
	 * sasl_client_init() function will do that for us (and return an error
	 * if it encounters something that it doesn't like).
	 */
	php_sasl_create_callbacks(Z_ARRVAL_P(array), SASLG(server_callbacks));

	if (sasl_server_init(SASLG(server_callbacks), name) != SASL_OK) {
		/* XXX: Free the callbacks? */
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */
/* {{{ proto resource sasl_server_new(string service [, string hostname [, string realm]])
   Creates a new SASL server connection context. */
PHP_FUNCTION(sasl_server_new)
{
	char *service, *hostname = NULL, *realm = NULL;
	int server_len, hostname_len, realm_len;
	sasl_conn_t *conn = NULL;
	int r = SASL_FAIL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|ss",
							  &service, &server_len,
							  &hostname, &hostname_len,
							  &realm, &realm_len) == FAILURE) {
		return;
	}

#if SASL_VERSION_MAJOR < 2
	r = sasl_server_new(service, hostname, realm, NULL, 0, &conn);
#else
	r = sasl_server_new(service, hostname, realm, NULL, NULL, NULL, 0, &conn);
#endif

	/* If sasl_server_new() failed, print an error and return false. */
	if (r != SASL_OK) {
		php_sasl_error(E_ERROR, r TSRMLS_CC);
		RETURN_FALSE;
	}

	ZEND_REGISTER_RESOURCE(return_value, conn, le_conn);
}
/* }}} */
/* {{{ proto int sasl_server_start(resource conn, string mech, string input, string &$output)
   Begins the auhentication process with a client. */
PHP_FUNCTION(sasl_server_start)
{
	zval *rsrc, *output;
	sasl_conn_t *conn;
	char *mech, *input;
	int mech_len, input_len;
	php_sasl_char *data = NULL;
	unsigned data_len = 0;
	int r = SASL_FAIL;

	if (zend_parse_parameters(4 TSRMLS_CC, "rssz/",
							  &rsrc,
							  &mech, &mech_len,
							  &input, &input_len,
							  &output) == FAILURE) {
		return;
	}

	if (mech_len > SASL_MECHNAMEMAX) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING,
						 "mechanism name exceeds maximum length (%u chars)",
						 SASL_MECHNAMEMAX);
		RETURN_FALSE;
	}

	if (output) {
		zval_dtor(output);
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

#if SASL_VERSION_MAJOR < 2
	r = sasl_server_start(conn, mech, input, input_len, &data, &data_len, NULL);
#else
	r = sasl_server_start(conn, mech, input, input_len, &data, &data_len);
#endif

	/* Print a warning if we receive an unexpected result. */
	if ((r != SASL_OK) && (r != SASL_CONTINUE)) {
		php_sasl_error(E_WARNING, r TSRMLS_CC);
	}

	/* Store the output in the "output" parameter (passed by reference). */
	ZVAL_STRINGL(output, (char *)data, data_len, 1);
	PHP_SASL_FREE(data);

	/* Simply return the result code from sasl_server_start(). */
	RETURN_LONG(r);
}
/* }}} */
/* {{{ proto int sasl_server_step(resource conn, string input, string &$output)
   Performs a step of the authentication. */
PHP_FUNCTION(sasl_server_step)
{
	zval *rsrc, *output = NULL;
	sasl_conn_t *conn;
	char *input;
	int input_len;
	php_sasl_char *data = NULL;
	unsigned data_len = 0;
	int r = SASL_FAIL;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsz/",
							  &rsrc,
							  &input, &input_len,
							  &output) == FAILURE) {
		return;
	}

	if (output) {
		zval_dtor(output);
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

#if SASL_VERSION_MAJOR < 2
	r = sasl_server_step(conn, input, input_len, &data, &data_len, NULL);
#else
	r = sasl_server_step(conn, input, input_len, &data, &data_len);
#endif

	/* Print a warning if we receive an unexpected result. */
	if ((r != SASL_OK) && (r != SASL_CONTINUE)) {
		php_sasl_error(r, E_WARNING TSRMLS_CC);
	}

	/* Store the output in the "output" parameter (passed by reference). */
	ZVAL_STRINGL(output, (char *)data, data_len, 1);
	PHP_SASL_FREE(data);

	/* Simply return the result code from sasl_server_step(). */
	RETURN_LONG(r);
}
/* }}} */
/* {{{ proto string sasl_listmech(resource conn)
   Creates a string with a list of SASL mechanism support by the server. */
PHP_FUNCTION(sasl_listmech)
{
	zval *rsrc;
	sasl_conn_t *conn;
	php_sasl_char *data = NULL;
	unsigned data_len = 0, count;
	int r = SASL_FAIL;

	if (zend_parse_parameters(1 TSRMLS_CC, "r", &rsrc) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

	r = sasl_listmech(conn, NULL, NULL, " ", NULL, &data, &data_len, &count);

	/* If sasl_listmech() failed, print an error and return an empty string. */
	if (r != SASL_OK) {
		php_sasl_error(E_WARNING, r TSRMLS_CC);
		PHP_SASL_FREE(data);
		RETURN_EMPTY_STRING();
	}

	RETVAL_STRINGL((char *)data, data_len, 1);
	PHP_SASL_FREE(data);
}
/* }}} */
/* {{{ proto bool sasl_checkpass(resource conn, string user, string pass)
   Checks a plaintext password for the given user. */
PHP_FUNCTION(sasl_checkpass)
{
	zval *rsrc;
	sasl_conn_t *conn;
	char *user, *pass;
	int user_len, pass_len;
	int r = SASL_FAIL;

	if (zend_parse_parameters(3 TSRMLS_CC, "rss",
							  &rsrc,
							  &user, &user_len,
							  &pass, &pass_len) == FAILURE) {
		return;
	}

	ZEND_FETCH_RESOURCE(conn, sasl_conn_t *, &rsrc, -1, le_conn_name, le_conn);

#if SASL_VERSION_MAJOR < 2
	r = sasl_checkpass(conn, user, user_len, pass, pass_len, NULL);
#else
	r = sasl_checkpass(conn, user, user_len, pass, pass_len);
#endif

	if (r != SASL_OK) {
		RETURN_FALSE;
	}

	RETURN_TRUE;
}
/* }}} */

/*
 * Local variables:
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 * vim600: fdm=marker
 * vim: sw=4 ts=4 noet
 */
